﻿package {
	
	import flash.display.SimpleButton;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.display.MovieClip;
	import flash.display.Shape;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.ui.Keyboard;
	import flash.utils.Timer;
	
	
	/**
	 * 1. フィールドに壁を描画する
	 * 2. フィールドにテトリミノを表示してみる
	 * 3. テトリミノを動かしてみる
	 * 4. テトリミノの回転
	 * 5. 回転の補正
	 * 6. テトリミノを固定する
	 * 7. ラインを消す
	 * 8. システム1（ゲームスタート、ゲームオーバー、ボタンクラスの機能拡張）
	 * 9. システム2（テトリミノ自動落下機能追加）
	 * 
	 * @author necoEngine bamboo-satellites
	 * @version 0.1.0
	 * @see http://www.bamboo-satellites.com/
	 * 
	 * コードブロック
	 * Constructor
	 * - Init
	 * - System
	 * - Tetromino Create
	 * - Tetromino Control
	 * - Field Operater
	 */
	public class main extends MovieClip {
		
		private const CELL_SIZE:uint      = 20;   // セルのサイズ。正方形
		private const FIELD_LEN_W:uint    = 12;   // フィールドの完全幅。壁を除くと10
		private const FIELD_LEN_H:uint    = 23;   // フィールドの完全高さ。壁、テトリミノ出現領域分を除くと20
		private const NUM_TETROMINOS:uint = 7;    // テトリミノの種類数
		
		private const LINE_COLOR:uint = 0x999999; // セルの線の色
		private const FILL_RAG:uint   = 120;      // ラインを消すときのタイムラグ
		private const END_DELEY:uint  = 30;       // ゲームオーバー時の描画間隔
		
		private const DROP_MAX:uint   = 1000;     // テトリミノが落下する速度の最低値（deleyプロパティの最大値）
		private const DROP_MIN:uint   = 50;       // テトリミノが落下する速度の最高値（deleyプロパティの最小値）
		private const ACCEL_MAX:uint  = 10;       // テトリミノが落下する速度の加速度上限（dropAccel最大値）
		private const BOOST:uint      = 10;       // dropCountの上限
		
		private const // フィールドデータで使用するコード
			EMPTY:uint      = 0,   // 空白コード
			TETROMINO1:uint = 1,   // テトリミノコード 棒形
			TETROMINO2:uint = 2,   // テトリミノコード　正方形
			TETROMINO3:uint = 3,   // テトリミノコード　S字
			TETROMINO4:uint = 4,   // テトリミノコード　Z字
			TETROMINO5:uint = 5,   // テトリミノコード　J字
			TETROMINO6:uint = 6,   // テトリミノコード　L字
			TETROMINO7:uint = 7,   // テトリミノコード　T字
			FIXED:uint      = 10,  // テトリミノ固定コード
			WALL:uint       = 99,  // 壁コード
			GAME_OVER:uint  = 100; // ゲームオーバーコード
		
		private const // 描画メソッドに使用する描画タイプ指示コード
			DROP:uint  = 0, // 落下中コード
			CLEAR:uint = 1, // 削除指示コード
			FILL:uint  = 2; // ラインを更新する指示コード
		
		
		private var fieldData:Vector.<Vector.<int>>;      // フィールドの状態を常に保存しておく配列
		private var fieldContainer:Shape;                 // フィールドのデータを元に状態を描画する表示コンテナ
		private var dropTetromino:Vector.<Vector.<uint>>; // 落下中のテトリミノのパターンデータ
		private var dropLocation:Point;                   // 落下中のテトリミノの位置情報
		
		private var dropTimer:Timer;                      // テトリミノの自動落下用タイマーオブジェクト
		private var dropAccel:int;                        // テトリミノの自動落下の加速値（deleyプロパティから減算する値）
		private var dropCount:uint;                       // テトリミノが落下した回数を数えるカウンタ（加速度の更新に使用）
		
		private var startButton:ButtonCreater;            // スタートボタンインスタンス
		
		
		public function main():void {
			
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
			
		} // constructor close
		
		
		
//================================================================================
// Init
//
// データの初期化関連の処理
//================================================================================
		private function init(e:Event = null):void {
			
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			
			// フィールドデータの初期化
			fieldData = fieldInit();
			
			// フィールドの状態を描画する表示コンテナを生成して、配置。
			fieldContainer = new Shape();
			addChild(fieldContainer);
			fieldContainer.x = CELL_SIZE;
			fieldContainer.y = CELL_SIZE;
			// フィールド描画。壁だけが描画される
			drawField(fieldData, fieldContainer);
			
			//スタートボタン設定
			startButton = new ButtonCreater(" S T A R T ");
			addChild(startButton);
			startButton.x = stage.stageWidth  - startButton.width  - CELL_SIZE;
			startButton.y = stage.stageHeight - startButton.height - CELL_SIZE;
			startButton.addEventListener(MouseEvent.MOUSE_DOWN, startHandler);
			
		} // init() close
		
		
		private function fieldInit():Vector.<Vector.<int>> {
			/**
			 * 機能：フィールドの初期化データ配列を返す
			 * 引数：なし
			 * 戻り値：Vector2次元配列
			 * 備考：最終的には初期描画処理も含める予定
			 */
			
			return Vector.<Vector.<int>>([
				Vector.<int>([ WALL, WALL, WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL, WALL, WALL]),
				Vector.<int>([ WALL, WALL, WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL, WALL, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL])
			]);
			
		} // fieldInit() close
		
		private function tetriminoInit(code:uint):Vector.<Vector.<uint>> {
			/**
			 * 機能：引数で取ったテトリミノコードに対応したテトリミノのパターン配列を返す
			 * 引数：uint テトリミノのタイプを表すコード
			 * 戻り値：Vector2次元配列　テトリミノのパターンデータ。
			 * 備考：
			 */
			
			switch (code) {
				case TETROMINO1: // 棒形
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY]),
						Vector.<uint>([TETROMINO1,TETROMINO1,TETROMINO1,TETROMINO1]),
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY]),
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY])
					]);
				case TETROMINO2: // 正方形
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO2, TETROMINO2]),
						Vector.<uint>([TETROMINO2, TETROMINO2])
					]);
				case TETROMINO3: // S字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      TETROMINO3, TETROMINO3]),
						Vector.<uint>([TETROMINO3, TETROMINO3, EMPTY]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO4: // Z字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO4, TETROMINO4, EMPTY]),
						Vector.<uint>([EMPTY,      TETROMINO4, TETROMINO4]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO5: // J字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO5, EMPTY,      EMPTY]),
						Vector.<uint>([TETROMINO5, TETROMINO5, TETROMINO5]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO6: // L字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      EMPTY,      TETROMINO6]),
						Vector.<uint>([TETROMINO6, TETROMINO6, TETROMINO6]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				default: // T字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      TETROMINO7, EMPTY]),
						Vector.<uint>([TETROMINO7, TETROMINO7, TETROMINO7]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
			}
			
		} // tetriminoInit() close
		
		private function colorInit(code:uint):uint {
			/**
			 * 機能：引数で取ったコードに対応したカラー値を返す
			 * 引数：uint
			 * 戻り値：uint 16進数
			 * 備考：
			 */
			
			switch (code) {
				case TETROMINO1: return 0x4DE6E1; // 棒形
				case TETROMINO2: return 0xE9EE11; // 正方形
				case TETROMINO3: return 0x46DF20; // S字
				case TETROMINO4: return 0xED2212; // Z字
				case TETROMINO5: return 0x2746D8; // J字
				case TETROMINO6: return 0xE49E1B; // L字
				case TETROMINO7: return 0xEA68E7; // T字
				case GAME_OVER:  return 0x666666; // ゲームオーバー描画カラー
				default:         return 0x7A5B52; // 壁
			}
			
		} // colorInit() close
		
		private function pointInit(code:uint):uint {
			/**
			 * 機能：引数で取ったコードに対応した位置座標の初期配置x値を返す
			 * 引数：uint
			 * 戻り値：uint
			 * 備考：
			 */
			
			switch (code) {
				case TETROMINO2: return 5;
				case TETROMINO4: return 5;
				case TETROMINO6: return 5;
				default:         return 4;
			}
			
		} // pointInit() close
		
		
		
//================================================================================
// System
//
// 主にゲームを運行して行く為の大枠になる処理
//================================================================================
		private function startHandler(event:MouseEvent):void {
			/**
			 * 機能：一番最初のゲームスタート処理を行う（再スタートはrestartHandler()）リスナー関数
			 * 引数：イベントオブジェクト
			 * 戻り値：なし
			 * 備考：一度しか実行されない。
			 */
			
			//スタートボタンのイベントリスナー削除
			startButton.removeEventListener(MouseEvent.MOUSE_DOWN, startHandler);
			startButton.activeShifter(false);
			
			// テトリミノデータを用意してフィールドデータに反映して描画
			dropLocation  = new Point(0, 0);
			dropTetromino = new Vector.<Vector.<uint>>(4);
			setTetromino(fieldData, dropTetromino, dropLocation, true);
			fieldMapping(fieldData, dropTetromino, dropLocation);
			drawField(fieldData, fieldContainer);
			
			// テトリミノ自動落下タイマー設定
			dropAccel = 1;
			dropTimer = new Timer(DROP_MAX);
			dropTimer.addEventListener(TimerEvent.TIMER, dropTimerHandler);
			dropTimer.start();
			
			// ステージイベント
			stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
			
		} // gameStartHandler() close
		
		
		private function restartHandler(event:MouseEvent):void {
			/**
			 * 機能：再スタート用リスナー関数。フィールドデータをクリアして新しいテトリミノのデータをマッピング、再描画までを実行する。
			 * 引数：イベントオブジェクト
			 * 戻り値：なし
			 * 備考：気になるので、使っていたの参照は一度切る方針。
			 */
			
			//スタートボタンのイベントリスナー削除
			startButton.removeEventListener(MouseEvent.MOUSE_DOWN, restartHandler);
			startButton.activeShifter(false);
			
			// fieldData配列の初期化
			fieldData.forEach(function (element:Vector.<int>, index:int, array:Vector.<Vector.<int>>):void { element = null; } );
			fieldData = null;
			fieldData = fieldInit();
			
			// 位置情報の初期化
			dropLocation = null;
			dropLocation = new Point(0, 0);
			
			// テトリミノパターンデータ配列の初期化
			dropTetromino.forEach(function (element:Vector.<uint>, index:int, array:Vector.<Vector.<uint>>):void { element = null;} );
			dropTetromino = null;
			dropTetromino = new Vector.<Vector.<uint>>(4);
			
			// テトリミノデータを用意してフィールドデータに反映
			setTetromino(fieldData, dropTetromino, dropLocation, true);
			fieldMapping(fieldData, dropTetromino, dropLocation);
			drawField(fieldData, fieldContainer);
			
			// 自動落下タイマーオブジェクトの初期化
			dropAccel = 1;
			dropTimer = new Timer(DROP_MAX);
			dropTimer.addEventListener(TimerEvent.TIMER, dropTimerHandler);
			dropTimer.start();
			
			//　ステージイベント
			stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
			
		} // restartHandler() close
		
		
		private function onGameOver(data:Vector.<Vector.<int>>, container:Shape):void {
			/**
			 * 機能：ゲームオーバー用描画のループ処理を行う。
			 * 引数：第1引数 フィールドデータ
			 *      第2引数 フィールドを描画する表示コンテナ
			 * 戻り値：なし
			 * 備考：具体的にはゲームオーバー時用のフィールドマッピングと描画実行、スタートボタンのリスナー関数登録
			 */
			
			var x:int, y:int, index:int = FIELD_LEN_H - 1, mappingTimer:Timer;
			
			mappingTimer = new Timer(END_DELEY);
			mappingTimer.addEventListener(TimerEvent.TIMER,
				function (event:TimerEvent):void { //　無名リスナー関数の登録
					
					for (x = 0; x < FIELD_LEN_W; x++) {
						if (data[index][x] > EMPTY) data[index][x] = GAME_OVER;
						drawField(fieldData, fieldContainer);
					}
					
					index--;
					
					if (index < 0) {
						mappingTimer.stop();
						mappingTimer.removeEventListener(TimerEvent.TIMER, arguments.callee);
						
						// スタートボタンインスタンスに再スタートのリスナー関数を登録する。
						startButton.addEventListener(MouseEvent.MOUSE_DOWN, restartHandler);
						startButton.activeShifter();
						
						mappingTimer = null;
					}
				}
			);
			mappingTimer.start();
			
		} // onGameOver() close
		
		
		private function dropTimerHandler(event:TimerEvent):void {
			/**
			 * 機能：一定時間毎にテトリミノを下へ移動する
			 * 引数：イベントオブジェクト
			 * 戻り値：なし
			 * 備考：
			 */
			
			tetrominoControl(Keyboard.DOWN,  fieldData, dropTetromino, dropLocation, fieldContainer);
			
		} // dropTimerHandler() close
		
		
		private function dropTimerRegulator(
				timer:Timer,         // 自動落下用のタイマーオブジェクト
				accel:int,           // 加速値
				counter:uint,        // 固定テトリミノカウンタ
				slow:Boolean = false // 減速指定かどうかの指示。減速処理なら真。
			):void {
			/**
			 * 機能：テトリミノの自動落下の時間間隔を調整する。
			 *      通常はテトリミノが一つ固定される度に自動落下の時間間隔値を減算していく（落下が早くなる）。
			 *      時間間隔値から減算する値（accel）は、固定されたテトリミノが一定数に到達すると増加して行く。
			 *      テトリスで消した場合は自動落下の時間間隔値が増加（落下が遅くなる）、減算値は減少する。
			 * 引数：宣言部に記載
			 * 戻り値：なし
			 * 備考：
			 */
			
			if (slow) {
				// 加速度の減少
				accel -= 2;
				if (accel < 1) accel = 1;
				// 落下間隔を増加
				timer.delay *= 2;
				if (timer.delay > DROP_MAX) timer.delay = DROP_MAX;
				
			} else {
				// 加速度の処理
				if (counter >= BOOST) {
					counter = 0; // リセット
					if (accel < ACCEL_MAX) accel++;
				} else {
					counter++; // 固定したテトリミノをカウント
				}
				
				// 自動落下速度の処理
				if (timer.delay <= DROP_MIN) timer.delay = DROP_MIN;
				else timer.delay -= accel * 2;
			}
			
		} // dropTimerRegulator() close
		
		
		
//================================================================================
// Tetromino Create
//
// 主に新しいテトリミノパターンデータの作成にかかわる処理
//================================================================================
		private function setTetromino(
				data:Vector.<Vector.<int>>,       // フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // テトリミノパターンデータ配列
				loc:Point,                        // テトリミノの位置情報Point
				first:Boolean = false             // 一番最初の配置かどうかの真偽値。
			):Boolean {
			/**
			 * 機能：新しいテトリミノパターンデータを用意する。フィールドデータには反映せず、あくまで用意するのみ。
			 * 引数：宣言部に記載。
			 *      第4引数について 真の場合一番最初として扱う。省略可で、省略した場合は一番最初ではない扱い。
			 * 戻り値：なし
			 * 備考：落下中のテトリミノパターンデータ配列長を新しく用意したパターンデータ配列長とそろえる必要がある
			 */
			
			var x:uint, y:uint, len:uint, gap:int, code:int, patternBuff:Vector.<Vector.<uint>>;
			
			code = Math.floor(Math.random() * 7) + 1; // 新しいパターンコード
			patternBuff = tetriminoInit(code);        // 新しいパターンデータ一時保存
			len = patternBuff.length;                 // 新しいパターンデータ配列長
			
			gap = tetromino.length - len; // 現在のdropTetrominoの一次元目の配列長との差
			if (gap > 0) for (y = gap;      y > 0; y--) tetromino.pop();                        // 一次元目減少
			if (gap < 0) for (y = gap * -1; y > 0; y--) tetromino.push(new Vector.<uint>(len)); // 一次元目増加
			
			for (y = 0; y < len; y++) {
				if (first) {
					tetromino[y] = new Vector.<uint>(len); // フィールドデータ初期化時の一番最初の実行
				} else {
					gap = tetromino[y].length - len; // 現在のdropTetrominoの二次元目の配列長との差
					if (gap > 0) for (x = gap;      x > 0; x--) tetromino[y].pop();   // 二次元目減少
					if (gap < 0) for (x = gap * -1; x > 0; x--) tetromino[y].push(0); // 二次元目増加
				}
				
				tetromino[y] = patternBuff[y].concat();
			}
			
			// 初期配置座標設定
			loc.x = pointInit(code - 1);
			loc.y = 0;
			
			// ゲームオーバー判定の結果を返す。ゲームオーバーじゃなければ真、ゲームオーバーなら偽
			return (!first && overlapChecker(data, tetromino, loc))? false: true;
			
		} // setTetromino() close
		
		
		
//================================================================================
// Tetromino Control
//
// 主に落下してくるテトリミノの操作に関わる処理
//================================================================================
		private function onKeyDownHandler(event:KeyboardEvent):void {
			/**
			 * 機能：キーボートイベントを検出して、キーに割り当てられた処理へ移行する。
			 * 引数：イベント
			 * 戻り値：なし
			 * 備考：拡張性の維持の為にキーボードイベントの検出、振り分けの処理のみに限定する。それ以外の動作処理は記入しない。
			 */
			
			switch(event.keyCode) {
				case Keyboard.LEFT:  // 左
					tetrominoControl(Keyboard.LEFT,  fieldData, dropTetromino, dropLocation, fieldContainer); break;
				case Keyboard.RIGHT: // 右
					tetrominoControl(Keyboard.RIGHT, fieldData, dropTetromino, dropLocation, fieldContainer); break;
				case Keyboard.DOWN:  // 下
					tetrominoControl(Keyboard.DOWN,  fieldData, dropTetromino, dropLocation, fieldContainer); break;
				case Keyboard.UP:    // 回転
					tetrominoControl(Keyboard.UP,    fieldData, dropTetromino, dropLocation, fieldContainer); break;
			}
			
		} // onKeyDownHandler() close
		
		
		private function tetrominoControl(
				direct:uint,                      // テトリミノを動かす方向指示コード。基本的にはキーコード。
				data:Vector.<Vector.<int>>,       // フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // テトリミノパターンデータ
				loc:Point,                        // テトリミノ座標
				container:Shape                   // フィールド表示コンテナ
			):Boolean {
			/**
			 * 機能：指示に合わせてテトリミノを動かす。引数でとったコードに該当する処理を実行する。
			 * 引数：宣言部に記載
			 * 戻り値：移動方向に何かあるかどうかの真偽値。何かあった場合に真を返す。
			 * 備考：
			 */
			
			switch (direct) {
				case Keyboard.LEFT: // 左
					if (getLeftHit(data, tetromino, loc)) return true; // 真が返る（左に何かある）と処理を終了
					
					fieldMapping(data, tetromino, loc, CLEAR); // 動かせる場合はフィールドデータから直前のテトリミノデータを削除
					loc.x--;                                   // 位置情報の更新
					
					// 棒形縦、左壁際での補正。配列の縮小
					if (tetromino.length > 3 && loc.x < 0) sizeReduction(tetromino, loc);
					
					fieldMapping(data, tetromino, loc); // フィールドデータにテトリミノの新しい位置情報を反映
					drawField(data, container); // フィールドを再描画
					
					return false;
					
				case Keyboard.RIGHT: // 右
					if (getRightHit(data, tetromino, loc)) return true; // 真が返る（右に何かある）と処理を終了
					
					fieldMapping(data, tetromino, loc, CLEAR); // 動かせる場合はフィールドデータから直前のテトリミノデータを削除
					loc.x++;                                   // 位置情報の更新
					
					// 棒形縦、左壁際での補正。配列の復元
					if (tetromino.length > 3 && loc.x == 1) restoreSize(tetromino, loc);
					
					fieldMapping(data, tetromino, loc); // フィールドデータにテトリミノの新しい位置情報を反映
					drawField(data, container); // フィールドを再描画
					
					return false;
					
				case Keyboard.DOWN: // 下
					dropTimer.stop(); // 自動落下タイマーをいったん停止
					if (getBottomHit(data, tetromino, loc)) { // 下にぶつかった場合
						// 一時的にキー操作イベントを削除する
						stage.removeEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
						
						// 自動落下速度を調整
						dropTimerRegulator(dropTimer, dropAccel, dropCount);
						
						fieldMapping(data, tetromino, loc, FIXED);   // フィールドデータにテトリミノコードを固定する
						var lines:Vector.<uint> = lineChecker(data); // 揃っているラインがあるかどうかを問い合わせる
						
						// 揃ったラインがあった場合
						if (lines.length > 0) {
							// 揃っているラインを一時的に空白行にして描画する。
							fieldMapping(data, tetromino, loc, CLEAR, lines);
							drawField(data, container);
							
							// ラインが揃って空白を詰める処理を少し遅らせて実行
							var ragTimer:Timer = new Timer(FILL_RAG, 1);
							ragTimer.addEventListener(TimerEvent.TIMER,
								function(event:TimerEvent):void { // 無名関数の登録
									ragTimer.removeEventListener(TimerEvent.TIMER, arguments.callee);
									
									fieldMapping(data, tetromino, loc, FILL, lines); // 空白のラインを詰める
									baseClear(data, lines.length);                   // 固定フィールドより上の処理
									
									// 新しいテトリミノデータを用意
									if (setTetromino(data, tetromino, loc)) {
										// ゲームオーバーではない
										fieldMapping(data, tetromino, loc); // 新しいテトリミノをフィールドに反映
										drawField(data, container);         // フィールドの描画
										stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
										if (lines.length >= 4) dropTimerRegulator(dropTimer, dropAccel, dropCount, true);
										dropTimer.start();
									} else {
										// ゲームオーバー
										dropTimer.removeEventListener(TimerEvent.TIMER, dropTimerHandler);
										dropTimer = null;
										fieldMapping(data, tetromino, loc); // 重なった状態で反映される
										onGameOver(data, container);        // ゲームオーバー処理実行
									}
									
									ragTimer = null;
								});
							ragTimer.start();
							
							return true; // 処理終了
						}
						
						// 揃ったラインが無い場合。新しいテトリミノデータを用意
						if (setTetromino(data, tetromino, loc)) {
							// ゲームオーバーではない
							fieldMapping(data, tetromino, loc); // 新しいテトリミノをフィールドに反映
							drawField(data, container);         // フィールドの描画
							stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
							dropTimer.start();
						} else {
							// ゲームオーバー
							dropTimer.removeEventListener(TimerEvent.TIMER, dropTimerHandler);
							dropTimer = null;
							fieldMapping(data, tetromino, loc); // 重なった状態で反映される
							onGameOver(data, container);        // ゲームオーバー処理実行
						}
						
						return true; // 処理終了
					}
					
					// 動かせる場合
					fieldMapping(data, tetromino, loc, CLEAR); // フィールドデータから直前のテトリミノデータを削除
					loc.y++;                                   // 位置情報の更新
					fieldMapping(data, tetromino, loc);        // フィールドデータにテトリミノの新しい位置情報を反映
					drawField(data, container);                // フィールドを再描画
					dropTimer.start();
					
					return false;
					
				case Keyboard.UP: // 回転
					if (getRotateHit(data, tetromino, loc)) return true; // 真が返ると処理を終了
					
					fieldMapping(data, tetromino, loc, CLEAR); // 回転させる前に直前のテトリミノデータを削除
					
					// 回転補正
					if (loc.x == 0) {
						loc.x++; // loc.x:0の場合場合、loc.x++で確定。
						// 棒形例外用。復元処理。Pointはダミー。
						if (tetromino.length - tetromino[0].length != 0) restoreSize(tetromino, new Point(0, 0));
					}
					// loc.x+パターンデータ幅がフィールド幅と同じ場合、loc.x--で確定。
					if (loc.x + tetromino[0].length == FIELD_LEN_W) loc.x--;
					// loc.x+パターンデータ幅がフィールド幅より大きい場合、loc.x-=2で確定。棒形例外用。
					if (loc.x + tetromino[0].length > FIELD_LEN_W)  loc.x -= 2;
					
					tetrominoRotate(tetromino);         // テトリミノパターンデータを90度回転
					fieldMapping(data, tetromino, loc); // フィールドデータにテトリミノの新しい位置情報を反映
					drawField(data, container);         // フィールドを再描画
					
					return false;
			}
			
			return false;
			
		} // tetrominoControl() close
		
		
		//------------------------------------------------------------
		// Data Control Tools
		//
		// テトリミノのパターンデータ配列自体を操作する時に使う
		//------------------------------------------------------------
		private function tetrominoRotate(tetromino:Vector.<Vector.<uint>>):Vector.<Vector.<uint>> {
			/**
			 * 機能：テトリミノのパターンデータを90度回転する
			 * 引数：回転させたいテトリミノのパターンデータ配列
			 * 戻り値：回転させたパターンデータ配列
			 * 備考：
			 */
			
			var x:uint, y:uint,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length,
					patternBuff:Vector.<Vector.<uint>>;
			
			patternBuff = new Vector.<Vector.<uint>>(xSize);
			for (y = 0; y < ySize; y++) patternBuff[y] = tetromino[y].concat();
			for (y = 0; y < ySize; y++) {
				for (x = 0; x < xSize;  x++) tetromino[x][ySize - 1 - y] = patternBuff[y][x];
			}
			
			return tetromino;
			
		} // tetrominoRotate() close
		
		private function sizeReduction(tetromino:Vector.<Vector.<uint>>, loc:Point):Vector.<Vector.<uint>> {
			/**
			 * 機能：位置情報のx値が負になった場合に、
			 *      落下中のテトリミノパターンデータ配列の二次元目の配列長を要素一つ分縮小する。
			 * 引数：第1引数 縮小したいテトリミノのパターンデータ配列
			 *      第2引数 落下中のテトリミノの位置情報Point
			 * 戻り値：縮小したテトリミノのパターンデータ配列
			 * 備考：この処理への評価を通る状況は、
			 *      棒形テトリミノが特定の配列状態で、左の壁に接近した場合に限定される。restoreSize()と対と考える。
			 */
			
			tetromino.forEach(
				function (element:Vector.<uint>, index:int, array:Vector.<Vector.<uint>>):void {
					element.reverse(); element.pop();
				}
			);
			loc.x = 0;
			return tetromino;
			
		} // sizeReduction() close
		
		private function restoreSize(tetromino:Vector.<Vector.<uint>>, loc:Point):Vector.<Vector.<uint>> {
			/**
			 * 機能：落下中のテトリミノのパターンデータ配列の一次元目と二次元目の配列長が違っている場合にそれを同じ長さに復元する。
			 * 引数：復元したいテトリミノのパターンデータ配列
			 * 戻り値：復元したテトリミノのパターンデータ他配列
			 * 備考：この処理の評価を通る状況はsizeReduction()メソッドで縮小した次の動作で、
			 * 　　　　パターンデータ配列が右に動いた直後に限定される。sizeReduction()と対と考える。
			 */
			
			var xSize:uint = tetromino[0].length, ySize:uint = tetromino.length;
			if (ySize - xSize != 0) {
				tetromino.forEach(
					function (element:Vector.<uint>, index:int, array:Vector.<Vector.<uint>>):void {
						element.push(EMPTY); element.reverse();
					}
				);
				loc.x--;
			}
			return tetromino;
			
		} // restreSize() clsoe
		
		
		//------------------------------------------------------------
		// Tetromino Hit Check
		//
		// 機能：テトリミノパターン配列内のテトリミノコードが入った要素を探し、
		//      検出した場合はその要素のフィールドでの位置から隣接要素を参照して何かあれば真を返す。
		// 引数：宣言部に記載
		// 戻り値：問い合わせられた方向に対して何かを検出した場合は真を返す
		// 備考：壁もしくは固定されたテトリミノのコードを検出すると真を返して終了
		//------------------------------------------------------------
		private function getLeftHit(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 出現したテトリミノのパターンデータ配列
				loc:Point                         // Point 出現したテトリミノの位置
			):Boolean {
			// 左方向へのヒットチェック
			
			var x:int, y:int, fx:int, fy:int,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length;
			
			for (x = 0; x < xSize; x++) { // 左から右へ走査
			for (y = 0; y < ySize; y++) { // 上から下へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy][fx - 1] > FIXED) return true;
				}
			}
			}
			return false;
			
		} // getLeftHit() close
		
		private function getRightHit(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 出現したテトリミノのパターンデータ配列
				loc:Point                         // Point 出現したテトリミノの位置
			):Boolean {
			// 右方向へのヒットチェック
			
			var x:int, y:int, fx:int, fy:int,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length;
			
			for (x = xSize - 1; x >= 0; x--) { // 右から左へ走査
			for (y = 0; y < ySize; y++) {      // 上から下へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy][fx + 1] > FIXED) return true;
				}
			}
			}
			return false;
			
		} // getRightHit() close
		
		private function getBottomHit(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 出現したテトリミノのパターンデータ配列
				loc:Point                         // Point 出現したテトリミノの位置
			):Boolean {
			// 下方向へのヒットチェック
			
			var x:int, y:int, fx:int, fy:int,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length;
			
			for (y = ySize - 1; y >= 0; y--) { // 下から上へ走査
			for (x = 0; x < xSize; x++) {      // 左から右へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy + 1][fx] > FIXED) return true;
				}
			}
			}
			return false;
			
		} // getBottomHit() close
		
		private function getRotateHit(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 出現したテトリミノのパターンデータ配列
				loc:Point                         // Point 出現したテトリミノの位置
			):Boolean {
			/**
			 * 機能：テトリミノパターン配列のデータのコピーを90度回転させた上でテトリミノコードが入った要素を探し、
			 *      コードを検出した場合はその要素のフィールドでの位置の要素を参照して何かあれば真を返す。
			 * 引数：宣言部に記載
			 * 戻り値：真偽値
			 * 備考：回転させてシミュレーションを行う。この処理内ではあくまでもシミュレートに限定。
			 *      ホントのパターン配列、位置情報に対しては何も行わない事。
			 */
			
			var x:uint, y:uint, fx:uint, fy:uint, revise:int = 0,
					xSize:uint,
					ySize:uint = tetromino.length,
					patternBuff:Vector.<Vector.<uint>>;
			
			// 一度コピーを作成してから、パターンデータを90度回転させる
			patternBuff = new Vector.<Vector.<uint>>(xSize);
			for (y = 0; y < ySize; y++) patternBuff[y] = tetromino[y].concat();
			xSize = patternBuff[0].length;
			ySize = patternBuff.length;
			
			// 回転補正　左の壁際
			if (loc.x == 0) {
				revise = 1;
				// 棒形例外用。
				if (xSize - ySize != 0) {
					restoreSize(patternBuff, new Point(0, 0)); // 復元処理。ダミーのPointを渡す。
					xSize = patternBuff[0].length;
					ySize = patternBuff.length;
				}
			}
			// 回転補正 右の壁際
			if (loc.x + xSize == FIELD_LEN_W) revise = -1;
			if (loc.x + xSize > FIELD_LEN_W)  revise = -2; // 棒形例外用
			
			tetrominoRotate(patternBuff); // パターンデータを回転
			
			// フィールド上のセルを参照して何かと重なっていないかを調べて、戻り値（真偽値）をそのまま返す
			return overlapChecker(data, patternBuff, new Point(loc.x + revise, loc.y));
			
		} // getRotateHit() close
		
		
		
//================================================================================
// Field Operater
//
// 主にフィールドデータにかかわる処理や描画
//================================================================================
		private function fieldMapping(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 出現したテトリミノのパターンデータ配列
				loc:Point,                        // Point 出現したテトリミノの位置
				type:uint = DROP,                 // テトリミノコードをどう処理するかの処理指示コード
				lines:Vector.<uint> = null        // フィールドデータ上の、揃っているラインの行数番号の配列
			):void {
			/**
			 * 機能：フィールドデータ配列に、出現したテトリミノのパターンデータを反映する。
			 * 引数：宣言部に記載。下は第4引数の指示コードについて。
			 *      DROP:フィールドデータに反映 CLEAR:フィールドデータから削除
			 *      FIXED:フィールドデータに固定 FILL:揃ったラインを削除して詰める
			 * 戻り値：なし
			 * 備考：第5引数がnullじゃない時にtypeがFILL、CLEAR以外だとエラーになる。でもtryは付けないので注意
			 */
			
			var x:uint, y:uint, ex:uint, ey:uint, px:uint, py:uint, code:uint, index:uint,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length;
			
			// 揃ったラインのデータがあるかどうかで分岐
			if (lines) {
				ey = lines.length;
				for (y = 0; y < ey; y++) {
					if (type == CLEAR) { // 揃っているラインを空白データに差し替える
						data.splice(lines[y], 1, Vector.<int>([WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,WALL]));
					}
					if (type == FILL)  {
						data.splice(lines[y], 1); // 揃ったライン（空白）の配列データ削除
						// 固定フィールドの最上段に空白データを入れ込む
						data.splice(2, 0, Vector.<int>([WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,WALL]));
					}
				}
			} else {
				ex = loc.x + xSize; // x軸走査の終了位置
				ey = loc.y + ySize; // y軸走査の終了位置
				for (y = loc.y; y < ey; y++) {
				for (x = loc.x; x < ex; x++) {
					px = x - loc.x; // テトリミノパターンデータ配列用のインデックス
					py = y - loc.y; // テトリミノパターンデータ配列用のインデックス
					code = tetromino[py][px];
					if (code) {
						switch (type) {
							case CLEAR: data[y][x] = EMPTY;        break; // 削除
							case FIXED: data[y][x] = code + FIXED; break; // 固定
							default:    data[y][x] = code;                // 落下中
						}
					}
				}
				}
			}
			
		} // fieldMapping() close
		
		
		private function drawField(data:Vector.<Vector.<int>>, container:Shape):void {
			/**
			 * 機能：フィールドデータの情報に基ずいて、表示コンテナにフィールドを描画する。
			 * 引数：第1引数 フィールドデータ
			 *      第2引数 描画する表示コンテナ
			 * 戻り値：なし
			 * 備考：
			 */
			
			var x:int, y:int, px:int, py:int, code:int;
			
			container.graphics.clear();
			
			for (y = 0; y < FIELD_LEN_H; y++) {
				px = 0;
				for (x = 0; x < FIELD_LEN_W; x++) {
					code = data[y][x];
					if (code > EMPTY) {
						if (code < WALL && code > FIXED) code -= FIXED; // 固定コードの復元
						container.graphics.beginFill(colorInit(code));
						container.graphics.lineStyle(1, LINE_COLOR);
						container.graphics.drawRect(px, py, CELL_SIZE, CELL_SIZE);
						container.graphics.endFill();
					}
					px += CELL_SIZE;
				}
				py += CELL_SIZE;
			}
			
			container.graphics.endFill();
			
		} // drawField() close
		
		
		private function lineChecker(data:Vector.<Vector.<int>>):Vector.<uint> {
			/**
			 * 機能：フィールド上に揃ったラインがあるかどうかを探索し、揃った行があった場合はそのラインの行数を配列で返す
			 * 引数：フィールドデータ配列
			 * 戻り値：揃っているラインの行数情報を入れた配列
			 * 備考：
			 */
			
			var x:uint, y:int, emptyCount:int = 1, filled:Boolean, lines:Vector.<uint>;
			
			lines = new Vector.<uint>();
			
			for (y = FIELD_LEN_H - 2; y >= 2 && emptyCount > EMPTY; y--) {
				// 固定フィールドの下のラインから、空白のセルがあるかどうかを調べる
				filled = true;
				for (x = 0; x < FIELD_LEN_W && filled; x++) if (data[y][x] == EMPTY) filled = false;
				
				if (filled) {
					lines.push(y); // 空白の無いラインがあった場合は、そのラインのインデックスを保存
				} else {
					// ラインが揃っていないことがわかった場合は、そのラインが完全な空白行かどうかを調べる。
					// 空白行の場合はそれより上の行にテトリミノがある事がないので、探索ループを終了する。
					emptyCount = WALL * (-2); // 壁の値を減算しておく
					for (x = 0; x < FIELD_LEN_W; x++) emptyCount += data[y][x];
				}
			}
			
			return lines.reverse();
			
		} // lineChecker() close
		
		
		private function overlapChecker(
				data:Vector.<Vector.<int>>,       // Vector二次元配列 フィールドデータ配列
				tetromino:Vector.<Vector.<uint>>, // Vector二次元配列 調べたいテトリミノのパターンデータ配列
				loc:Point                         // テトリミノの座標
			):Boolean {
			/**
			 * 機能：新しく用意したテトリミノデータが、固定されたテトリミノと重なるかどうかを調べて真偽値を返す
			 * 引数：宣言部に記載
			 * 戻り値：固定されたテトリミノと重なり、ゲームオーバーだった場合は真を返す。
			 * 備考：
			 */
			
			var x:uint, y:uint, fx:uint, fy:uint,
					xSize:uint = tetromino[0].length,
					ySize:uint = tetromino.length;
			
			for (y = 0; y < ySize; y++) {
			for (x = 0; x < xSize; x++) {
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy][fx] > FIXED) return true;
				}
			}
			}
			return false;
			
		} // overlapChecker() close
		
		
		private function baseClear(data:Vector.<Vector.<int>>, len:uint):void {
			/**
			 * 機能：固定フィールドより上のセルに壁以外のコードが入っていた場合に、消えたラインの分だけ値を下へ移動する。
			 * 引数：第1引数 フィールドデータ配列、第2引数 揃ったラインの数
			 * 戻り値：なし
			 * 備考：やっつけである。参照が移ると困るので一度取り出してから。
			 */
			
			var x:uint, y:int, code:uint;
			for (y = 1; y >= 0; y--) {
			for (x = 3; x < 9;  x++) {
				if (data[y][x] < WALL) {
					code = data[y][x];
					data[y][x] = EMPTY;
					data[y + len][x] = code;
				}
			}
			}
			
		} // () close
	}
}